#pragma once

#include <CGAL/minkowski_sum_2.h>
#include <CGAL/Boolean_set_operations_2.h>
#include <CGAL/Small_side_angle_bisector_decomposition_2.h>

#include <CGAL/Arr_overlay.h>
#include <CGAL/Arr_default_overlay_traits.h>

#include <CGAL/Polygon_set_2.h>

#include <boost/bind.hpp>

/*!
 * Checks if two given polygons intersect in their interior. Uses boolean set-operations function.
 * \param P The first polygon.
 * \param Q The second polygon.
 * \return True if P and Q intersect in their interior, false otherwise.
 */
template <typename Polygon>
bool do_intersect_interior(const Polygon &P, const Polygon &Q)
{
	return CGAL::do_intersect(P, Q);
}

/*!
 * Checks if two given linear simple polygons intersect in their exterior.
 * \param P The first polygon.
 * \param Q The second polygon.
 * \return True if P and Q intersect in their exterior, false otherwise.
 */
template <class Kernel, class Container>
bool do_intersect_exterior(const CGAL::Polygon_2<Kernel, Container> &P, 
			 const CGAL::Polygon_2<Kernel, Container> &Q)
{
	CGAL_precondition(P.is_simple());
	CGAL_precondition(Q.is_simple());

	typedef CGAL::Polygon_set_2<Kernel, Container>								Polygon_set;
	typedef Polygon_set::Arrangement_2											Arrangement;

	/*!
	 * \struct
	 * This struct inherits from the default overlay traits and is used when checking if two polygons intersect.
	 * It forwards the calls to the default overlay traits when finding out that the boundary of the two polygons
	 * intersects it changes the value of the _do_intersect flag to true.
	 */
	struct Intersect_overaly_traits : public CGAL::Arr_default_overlay_traits<Arrangement>
	{
		typedef CGAL::Arr_default_overlay_traits<Arrangement>					Base;
		Intersect_overaly_traits() : do_intersects(false) {}

		void create_vertex(Vertex_handle_A v1, Vertex_handle_B v2, Vertex_handle_R v)
		{
			Base::create_vertex(v1, v2, v);
			do_intersects = true;
		}
		void create_vertex(Vertex_handle_A v1, Halfedge_handle_B e2, Vertex_handle_R v)
		{
			Base::create_vertex(v1, e2, v);
			do_intersects = true;
		}
		void create_vertex(Halfedge_handle_A e1, Vertex_handle_B v2, Vertex_handle_R v)
		{
			Base::create_vertex(e1, v2, v);
			do_intersects = true;
		}
		void create_vertex(Halfedge_handle_A e1, Halfedge_handle_B e2, Vertex_handle_R v)
		{
			Base::create_vertex(e1, e2, v);
			do_intersects = true;
		}
		// forwarding others
		void create_vertex(Vertex_handle_A v1, Face_handle_B f2, Vertex_handle_R v)
		{
			Base::create_vertex(v1, f2, v);
		}
		void create_vertex(Face_handle_A f1, Vertex_handle_B v2, Vertex_handle_R v)
		{
			Base::create_vertex(f1, v2, v);
		}

		bool do_intersects;
	};

	// In order to find out if two polygons intersect we overlay their underlying arrangements and use the ad-hoc overlay traits
	// above that will return us the right answer.
	Polygon_set ps_p(P);
	Polygon_set ps_q(Q);

	Arrangement arr_p = ps_p.arrangement();
	Arrangement arr_q = ps_q.arrangement();

	Arrangement overlay_arr;
	Intersect_overaly_traits ov_traits;
	CGAL::overlay (arr_p, arr_q, overlay_arr, ov_traits);
	return ov_traits.do_intersects;
}

/*!
 * Computes the overlap area of two linear simple polygons. Uses BSO and then add the area of the polygon and substract
 * the area of its holes.
 * \param P The first polygon.
 * \param Q The second polygon.
 * \return The area of the overlap between P and Q.
 */
template <class Kernel, class Container>
typename CGAL::Polygon_2<Kernel, Container>::FT
overlap_area(const CGAL::Polygon_2<Kernel, Container> &P, 
			 const CGAL::Polygon_2<Kernel, Container> &Q)
{
	CGAL_precondition(P.is_simple());
	CGAL_precondition(Q.is_simple());

	typedef typename CGAL::Polygon_2<Kernel, Container>::FT						FT;
	typedef CGAL::Polygon_with_holes_2<Kernel, Container>						Polygon_with_holes_2;
	typedef std::list<Polygon_with_holes_2>										Pol_list;
	Pol_list overlap;
	CGAL::intersection(P, Q, std::back_inserter(overlap));
	if (overlap.empty())
		return 0;
	
	// summing the areas and reducing the area of holes.
	FT result = 0;
	for (Pol_list::iterator it = overlap.begin(); it != overlap.end(); ++it)
	{
		Polygon_with_holes_2& cur = *it;
		result += cur.outer_boundary().area();
		for (Polygon_with_holes_2::Hole_const_iterator jt = cur.holes_begin();
			jt != cur.holes_end(); ++jt)
		{
			result -= jt->area();
		}
	}	

	return result;
}

/*!
 * Calculate the minimum vector in a certain direction a polygon Q needs to be translated in order not to 
 * interiorly intersect polygon P. If the polygons aren't intersecting in their interior then the 
 * vector (0, 0) is returned.
 * The function computes the minkowski sum of the two polygons, which can be quite heavy. If you (the user)
 * computed already the minkowski sum then you can use this function as a reference on how to compute the 
 * minimal vector.
 * \param P The first polygon.
 * \param Q The second polygon.
 * \param d The direction in which we want to move Q.
 * \return The minimal vector in direction d that translating Q by it will cause P and Q not to intersect.
 */
template <class Kernel, class Container>
typename Kernel::Vector_2
minimal_translation(const CGAL::Polygon_2<Kernel, Container> &P, 
					const CGAL::Polygon_2<Kernel, Container> &Q, 
					const typename Kernel::Direction_2 &d)
{
	typedef CGAL::Polygon_2<Kernel, Container>						Polygon;
	typedef typename Polygon::Segment_2								Segment;
	typedef CGAL::Aff_transformation_2<Kernel>						Transform;
	typedef typename Kernel::Direction_2							Direction;
	typedef typename Kernel::Ray_2									Ray;
	typedef typename Kernel::Point_2								Point;
	typedef typename Kernel::FT										FT;
	typedef CGAL::Polygon_with_holes_2<Kernel, Container>			Polygon_with_holes;
	typedef std::list<Segment>										Segments_list;
	typedef std::vector<Point>										Points_vector;

	// constructing a ray in the wanted direction. We'll intersect that ray with the minkowski sum
	// to find the possible vectors. Then we take the minimal vector.
	Kernel k;
	Point p = k.construct_point_2_object()(CGAL::ORIGIN);
	Ray r = k.construct_ray_2_object()(p, d);

	// if the polygons don't intersect we can return (0, 0).
	if (do_intersect_interior(P, Q) == false)
		return k.construct_vector_2_object()(CGAL::ORIGIN, p);

	// the minimum translation needed is the minimum translation needed in order to move the 
	// minkowski sum to sum position that it won't contain the origin. we compute the minkowski
	// sum of the two polygons and intersect it with a ray in the required direction. We take the
	// point which is closest to the origin.
	Polygon Q_tag = transform(Transform(CGAL::Rotation(), 0, -1), Q);
	CGAL::Small_side_angle_bisector_decomposition_2<Kernel>  ssab_decomp;
	Polygon_with_holes sum = minkowski_sum_2 (P, Q_tag, ssab_decomp);
	CGAL_assertion(CGAL::bounded_side_2(sum.outer_boundary().vertices_begin(), 
		sum.outer_boundary().vertices_end(), p) == CGAL::ON_BOUNDED_SIDE);
	
	// get all the edges.
	Segments_list segments;
	std::copy(sum.outer_boundary().edges_begin(), sum.outer_boundary().edges_end(), std::back_inserter(segments));
	for (Polygon_with_holes::Hole_const_iterator it = sum.holes_begin();
		it != sum.holes_end(); ++it)
	{
		std::copy(it->edges_begin(), it->edges_end(), std::back_inserter(segments));
	}

	// get intersection points in the wanted direction.
	Points_vector points;
	for (Segments_list::iterator it = segments.begin(); it != segments.end(); ++it)
	{
		CGAL::Object obj = CGAL::intersection(r, *it);
		Point ip;
		if (CGAL::assign(ip, obj)) 
		{
			points.push_back(ip);
		}
	}
	CGAL_assertion(points.empty() == false);

	// These two lines return the point which is the closest to the origin. It uses Boost.Bind and 
	// Boost.Lambda to generate the less prediate that we want. boost::bind(dist, p, _1) generates
	// a functor that computes the distance from the origin, and the outer bind statement compares
	// the two answers. If this is complicated you can just compute the distances and get the closest
	// point by yourselves.
	typename Kernel::Compute_squared_distance_2 dist = k.compute_squared_distance_2_object();
	Point min_point = *std::min_element(points.begin(), points.end(), 
		boost::bind(std::less<FT>(), boost::bind(dist, p, _1), boost::bind(dist, p, _2)));

	typename Kernel::Vector_2 v = k.construct_vector_2_object()(p, min_point);
	CGAL_postcondition(k.construct_direction_2_object()(v) == d);
	return v;
}

/*!
 * Calculate the minimum vector a polygon Q needs to be translated in order not to interiorly intersect
 * polygon P. If the polygons aren't intersecting in their interior then the vector (0, 0) is returned.
 * The function computes the minkowski sum of the two polygons, which can be quite heavy. If you (the user)
 * computed already the minkowski sum then you can use this function as a reference on how to compute the 
 * minimal vector.
 * \param P The first polygon.
 * \param Q The second polygon.
 * \return The minimal vector that translating Q by it will cause P and Q not to intersect.
 * \todo There is some repetition of code between this function and the previous function.
 */
template <class Kernel, class Container>
typename Kernel::Vector_2
minimal_translation(const CGAL::Polygon_2<Kernel, Container> &P, 
					const CGAL::Polygon_2<Kernel, Container> &Q)
{
	typedef CGAL::Polygon_2<Kernel, Container>						Polygon;
	typedef typename Polygon::Segment_2								Segment;
	typedef CGAL::Aff_transformation_2<Kernel>						Transform;
	typedef typename Kernel::Direction_2							Direction;
	typedef typename Kernel::Ray_2									Ray;
	typedef typename Kernel::Line_2									Line;
	typedef typename Kernel::Point_2								Point;
	typedef typename Kernel::FT										FT;
	typedef CGAL::Polygon_with_holes_2<Kernel, Container>			Polygon_with_holes;
	typedef std::list<Segment>										Segments_list;
	typedef std::vector<Point>										Points_vector;

	// constructing a ray in the wanted direction. We'll intersect that ray with the minkowski sum
	// to find the possible vectors. Then we take the minimal vector.
	Kernel k;
	Point p = k.construct_point_2_object()(CGAL::ORIGIN);

	// if the polygons don't intersect we can return (0, 0).
	if (do_intersect_interior(P, Q) == false)
		return k.construct_vector_2_object()(CGAL::ORIGIN, p);

	// the minimum translation needed is the minimum translation needed in order to move the 
	// minkowski sum to sum position that it won't contain the origin. we compute the minkowski
	// sum of the two polygons and intersect it with a ray in the required direction. We take the
	// point which is closest to the origin.
	Polygon Q_tag = transform(Transform(CGAL::Rotation(), 0, -1), Q);
	CGAL::Small_side_angle_bisector_decomposition_2<Kernel>  ssab_decomp;
	Polygon_with_holes sum = minkowski_sum_2 (P, Q_tag, ssab_decomp);
	CGAL_assertion(CGAL::bounded_side_2(sum.outer_boundary().vertices_begin(), 
		sum.outer_boundary().vertices_end(), p) == CGAL::ON_BOUNDED_SIDE);
	
	// get all the edges.
	Segments_list segments;
	std::copy(sum.outer_boundary().edges_begin(), sum.outer_boundary().edges_end(), std::back_inserter(segments));
	for (Polygon_with_holes::Hole_const_iterator it = sum.holes_begin();
		it != sum.holes_end(); ++it)
	{
		std::copy(it->edges_begin(), it->edges_end(), std::back_inserter(segments));
	}

	// get intersection points in the wanted direction.
	Points_vector points;
	for (Segments_list::iterator it = segments.begin(); it != segments.end(); ++it)
	{
		// we intersect with the perpendicular line that pass through the origin (this is how we get the closest point to the origin
		// that is on the segment. If there is no intersection with the segment, we take the end points.
		Line supp = k.construct_line_2_object()(*it);
		Line perpen = k.construct_perpendicular_line_2_object() (supp, p);
		
		CGAL::Object obj = CGAL::intersection(*it, perpen);
		Point ip;
		if (CGAL::assign(ip, obj)) 
		{
			points.push_back(ip);
		}
		else
		{
			points.push_back(k.construct_vertex_2_object()(*it, 0));
			points.push_back(k.construct_vertex_2_object()(*it, 1));
		}
	}
	CGAL_assertion(points.empty() == false);

	// These two lines return the point which is the closest to the origin. It uses Boost.Bind and 
	// Boost.Lambda to generate the less prediate that we want. boost::bind(dist, p, _1) generates
	// a functor that computes the distance from the origin, and the outer bind statement compares
	// the two answers. If this is complicated you can just compute the distances and get the closest
	// point by yourselves.
	typename Kernel::Compute_squared_distance_2 dist = k.compute_squared_distance_2_object();
	Point min_point = *std::min_element(points.begin(), points.end(), 
		boost::bind(std::less<FT>(), boost::bind(dist, p, _1), boost::bind(dist, p, _2)));

	typename Kernel::Vector_2 v = k.construct_vector_2_object()(p, min_point);
	return v;
}

